
// ****************************************************************************
//                                 
//                                 LC Meter
//
// ****************************************************************************

#include "include.h"

// Resonator: f=1/(2*pi*sqrt(L*C))
// check C: L = 100uH, parallel C = 1nF
// C range: 0.1pF ... 1uF
//   C = (1/(f*2*pi))^2 / L0 - C0
// check L: serial L = 100uH, parallel C = 1nF
// L range: 10nH ... 100mH
//   L = (1/(f*2*pi))^2 / C0 - L0
// frequency: 15 kHz ... 530 kHz

// Timer1 overflow counter (after 65K samples)
volatile u16 LC_Over = 0;

// last counter value (value in range 3000..150000)
volatile u32 LC_Counter = 0;

// time period in us of last counter value (approx. 250000)
volatile u32 LC_CounterTim = 250000;

// time of last read
u16 LC_Time = 0;
u32 LC_Last = 0;

// flag - reading L instead of C
//Bool LC_LMode = False;

// time of previous mode
//u32 LC_ModeTime = 0;

// measured C in 0.1pF (scale = -13)
//s32 LC_C = 0;

// measured L in 10nH (scale = -8)
s32 LC_L = 0;

// LC frequency
u32 LC_Freq = 0;

// reference C in 0.1pF (scale = -13)
u32 C0Ref = C0REF; // 1nF

// reference L in 10nH (scale = -8)
u32 L0Ref = L0REF; // 68uH

// tare C
//s32 CTare = 0;

// tare L
s32 LTare = L_TARE;

// external interrupt request of Timer1 overflow
ISR(TIMER1_OVF_vect)
{
	LC_Over++;
}

// initialize LC meter
void LC_Init()
{
	// Timer 1 initializazion: clock source T1 pin rising edge
	TCCR1B = (1 << CS12) | (1 << CS11) | (1 << CS10);
	TCNT1 = 0;
	TIMSK |= (1 << TOIE1); // interrupt on overflow

	// pull-up on LC mode input
	//PORTB |= B3;

	// last mode change time (interval 10 secs)
	//LC_ModeTime = GetTime() - 10000000;
}

// update LC counter value (called from timer every 250ms, with interrupts disabled)
// tim = delta time in us
void LC_Update(u32 tim)
{
	u16 hl = TCNT1;
	TCNT1 = 0;

	u16* v = (u16*)&LC_Counter;
	v[0] = hl;
	v[1] = LC_Over;

	LC_CounterTim = tim;
	LC_Over = 0;
}

#ifndef DO_TEST // do some test (reduce programs)

// read LC values
void LC_Read()
{
	// read LC counter
	u32 cnt;
	u32 tim;
	cli();
	cnt = LC_Counter; // value in range 3000..150000
	tim = LC_CounterTim; // value approx. 250000
	sei();
	LC_Freq = cnt * 1000 / (tim/10) * 100; // frequency in Hz

	// get LC mode (L or C)
/*	Bool mode = ((PIND & B2) == 0);
	u32 t = GetTime();
	if (mode == LC_LMode) // mode not changed
	{
		if (t - LC_ModeTime > 10000000)
			LC_ModeTime = t - 10000000; // hold max. delta time
	}
	else
	{
		// mode changed
		LC_LMode = mode;

		// tare (if faster than 0.75 sec)
		if (t - LC_ModeTime < 750000) LC_DoTare = True;

		LC_ModeTime = t;
	}
*/
	u32 k;

	// calculate 1/((f*2*pi)^2)
	//   cnt = f in Hz (value 11k..560k)
	cnt = LC_Freq;
	if (cnt > 400000UL) // f=400..560 kHz
	{
		// quadrate frequency
		k=(cnt + 5)/10; // k=f/10, values 40k..56k
		k=(k*k + 20000UL)/40000UL; // k=(f^2)/4e6, values 40k..79k

		// reciprocal value
		k=(4000000000UL+k/2)/k; // 1/(f^2)*16e15, valules 100k..50k

		// multiply by 1/((2*pi)^2)
		k*=((253303UL+8)/16); // values 1.6G..790M

		// divide by reference, values 16k..8k
		//if (mode)
			k = (k + 5*C0Ref)/(10*C0Ref);
		//else
		//	k = (k + 5*L0Ref)/(10*L0Ref);
	}
	else if (cnt > 200000UL) // f=200..400 kHz
	{
		// quadrate frequency
		k=(cnt + 5)/10; // k=f/10, values 20k..40k
		k=(k*k + 5000UL)/10000UL; // k=(f^2)/1e6, values 40k..160k

		// reciprocal value
		k=(4000000000UL+k/2)/k; // 1/(f^2)*4e15, valules 100k..25k

		// multiply by 1/((2*pi)^2)
		k*=((25330UL+2)/4); // values 630M..158M

		// divide by reference, values 63k..16k
		//if (mode)
			k = (k + C0Ref/2)/C0Ref;
		//else
		//	k = (k + L0Ref/2)/L0Ref;
	}
	else if (cnt > 100000UL) // f=100..200 kHz
	{
		// quadrate frequency
		k=(cnt + 2)/4; // k=f/4, values 25k..50k
		k=(k*k + 10000UL)/20000UL; // k=(f^2)/32e4, values 31k..125k

		// reciprocal value
		k=(4000000000UL+k/2)/k; // 1/(f^2)*128e13, valules 129k..32k

		// multiply by 1/((2*pi)^2)
		k*=((2533030UL+64)/128); // values 2.5G..630M

		// divide by reference, values 250k..63k
		//if (mode)
			k = (k + C0Ref/2)/C0Ref;
		//else
		//	k = (k + L0Ref/2)/L0Ref;
	}
	else // f=11..100 kHz
	{
		// quadrate frequency
		k=(cnt + 1)/2; // k=f/2, values 6k..50k
		k=(k*k + 5000UL)/10000UL; // k=(f^2)/4e4, values 3k..250k

		// reciprocal value
		k=(4000000000UL+k/2)/k; // 1/(f^2)*16e13, valules 1.3M..16k

		// multiply by 1/((2*pi)^2)
		k*=((25330UL+8)/16); // values 2G..25M

		// divide by reference, values 20M..250k
		//if (mode)
			k = (k + C0Ref/200)/(C0Ref/100);
		//else
		//	k = (k + L0Ref/200)/(L0Ref/100);
	}

	// substract reference and tare
	//if (mode)
		LC_L = k - L0Ref - LTare;
	//else
	//	LC_C = k - C0Ref - CTare;
}

#endif // DO_TEST
