
// ****************************************************************************
//                                 
//  LCD display EL1602B (controller KS0066U, compatible with Hitachi HD44780)
//
// ****************************************************************************

#include "include.h"

// LCD pins:
//   PC0: D4 (data bit 4)
//   PC1: D5 (data bit 5)
//   PC2: D6 (data bit 6)
//   PC3: D7 (data bit 7)
//   PB0: RS (register select)
//   PB1: E (enable)

/*
CGROM (character generator ROM): 1260 bytes, 204 characters 5x8 and 32 characters 5x11
CGRAM (character generator RAM): 64 bytes, 8 characters 5x8 or 4 characters 5x11
DDRAM (display RAM): 80 bytes

2 rows with 16 characters

Pins and connection:

1 VSS (GND) .... GND (board 1)
2 VDD (+5 V) .... +5V (board 2 with +5V wire)
3 V0 (contrast adjustment)
4 RS (H/L register select signal) ... PC0 (CPU 23, board 3)
5 R/W (H/L Read/Write signal) ... PC1 (CPU 24, board 6)
6 E (H/L Enable signal)	... PC2 (CPU 25, board 4)
7 DB0 (H/L Data bus line) .... PB0 (CPU 12, board 12)
8 DB1 (H/L Data bus line) .... PB1 (CPU 13, board 13)
9 DB2 (H/L Data bus line) .... PB2 (CPU 14, board 14 - last)
10 DB3 (H/L Data bus line) .... PD3 (CPU 1, board 7)
11 DB4 (H/L Data bus line) .... PD4 (CPU 2, board 8)
12 DB5 (H/L Data bus line) .... PD5 (CPU 9, board 9)
13 DB6 (H/L Data bus line) .... PD6 (CPU 10, board 10)
14 DB7 (H/L Data bus line) .... PD7 (CPU 11, board 11)
15 K LED- (0V for back light)
16 A LED+ (+4.2V for back light)
*/

// LCD commands
#define LCDCMD_CLEAR	B0	// 00000001 Clear Display and Home the Cursor
#define LCDCMD_HOME	B1	// 0000001* Return Cursor and LCD to Home Position
#define LCDCMD_MOVE	B2	// 000001IS Set Cursor Move Direction (I: increment/decrement; S: shift display)
#define LCDCMD_DISPLAY	B3	// 00001DCB Enable Display/Cursor (D: display on, C: cursor on, B: blink on)
#define LCDCMD_SHIFT	B4	// 0001SR** Move Cursor/Shift Display (S: diplay shift, R: direction right)
#define LCDCMD_SETUP	B5	// 001DNF** Set Interface Length (D: 8 bit interface, N: 2 rows, F: big font 5x10)
#define LCDCMD_CGRAM	B6	// 01AAAAAA Move Cursor into CGRAM (A: address)
#define LCDCMD_DDRAM	B7	// 1AAAAAAA Move Cursor to Display (A: address)

// set ENABLE state ON
#define EN_ON PORTB |= B1

// set ENABLE state OFF
#define EN_OFF PORTB &= ~B1

// set DATA mode
#define DATA_MODE PORTB |= B0

// set COMMAND mode
#define CMD_MODE PORTB &= ~B0

// Custom font characters (every byte is 1 row of characters; only bits 0..4 are used)
const u8 Font5x7[CHAR_NUM*8] EEMEM = {

	// diode 1: (A) -|>|- (C)
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 1 0 0 1
	// 0 0 0 1 1 1 0 1
	// 0 0 0 1 1 1 1 1
	// 0 0 0 1 1 1 0 1
	// 0 0 0 1 1 0 0 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 0 0 0 0 0
	0x11, 0x19, 0x1d, 0x1f, 0x1d, 0x19, 0x11, 0x00,

	// capacitor EXCELENT: -||-
	// 0 0 0 0 0 1 0 0
	// 0 0 0 1 1 1 1 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 1 1 1 1
	// 0 0 0 0 0 0 0 0
	// 0 0 0 1 1 1 1 1
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 1 0 0
        0x04, 0x1f, 0x11, 0x1f, 0x00, 0x1f, 0x04, 0x04,

	// capacitor GOOD: -||-
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 1 0 0
	// 0 0 0 1 1 1 1 1
	// 0 0 0 0 0 0 0 0
	// 0 0 0 1 1 1 1 1
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 0 0 0
        0x04, 0x04, 0x1f, 0x00, 0x1f, 0x04, 0x04, 0x00,

	// capacitor BAD: -||-
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 1 0 0
	// 0 0 0 1 0 1 0 1
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 1 0 1 0 1
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 1 0 0
        0x04, 0x04, 0x15, 0x00, 0x00, 0x15, 0x04, 0x04,

	// resistor: -[]-
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 1 1 1 0
	// 0 0 0 0 1 0 1 0
	// 0 0 0 0 1 0 1 0
	// 0 0 0 0 1 0 1 0
	// 0 0 0 0 1 1 1 0
	// 0 0 0 0 0 1 0 0
	// 0 0 0 0 0 0 0 0
//        0x04, 0x0e, 0x0a, 0x0a, 0x0a, 0x0e, 0x04, 0x00,

/*
	// diode 2: (C) -|<|- (A)
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 0 1 1 1
	// 0 0 0 1 0 1 1 1
	// 0 0 0 1 1 1 1 1
	// 0 0 0 1 0 1 1 1
	// 0 0 0 1 0 0 1 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 0 0 0 0 0
	0x11, 0x13, 0x17, 0x1f, 0x17, 0x13, 0x11, 0x00,

	// capacitor: -||-
	// 0 0 0 1 1 0 1 1
	// 0 0 0 1 1 0 1 1
	// 0 0 0 1 1 0 1 1
	// 0 0 0 1 1 0 1 1
	// 0 0 0 1 1 0 1 1
	// 0 0 0 1 1 0 1 1
	// 0 0 0 1 1 0 1 1
	// 0 0 0 0 0 0 0 0
        0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x00,

	// resistor left: -[
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 1 1 1 1
	// 0 0 0 0 1 0 0 0
	// 0 0 0 1 1 0 0 0
	// 0 0 0 0 1 0 0 0
	// 0 0 0 0 1 1 1 1
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
        0x00, 0x0f, 0x08, 0x18, 0x08, 0x0f, 0x00, 0x00,

	// resistor right: ]-
	// 0 0 0 0 0 0 0 0
	// 0 0 0 1 1 1 1 0
	// 0 0 0 0 0 0 1 0
	// 0 0 0 0 0 0 1 1
	// 0 0 0 0 0 0 1 0
	// 0 0 0 1 1 1 1 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
        0x00, 0x1e, 0x02, 0x03, 0x02, 0x1e, 0x00, 0x00,

	// coil left: ^^
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 1 1 0 1
	// 0 0 0 1 0 0 1 0
	// 0 0 0 1 0 0 1 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
        0x00, 0x00, 0x0d, 0x12, 0x12, 0x00, 0x00, 0x00,

	// coil right: ^^
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 1 0 1 1 0
	// 0 0 0 0 1 0 0 1
	// 0 0 0 0 1 0 0 1
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 0 0 0 0
        0x00, 0x00, 0x16, 0x09, 0x09, 0x00, 0x00, 0x00,

	// transistor
	// 0 0 0 0 0 0 0 1
	// 0 0 0 0 0 1 0 1
	// 0 0 0 0 0 1 1 0
	// 0 0 0 1 1 1 0 0
	// 0 0 0 0 0 1 1 0
	// 0 0 0 0 0 1 0 1
	// 0 0 0 0 0 0 0 1
	// 0 0 0 0 0 0 0 0
        0x01, 0x05, 0x06, 0x1c, 0x06, 0x05, 0x01, 0x00,
*/
	// omega
	// 0 0 0 0 0 0 0 0
	// 0 0 0 0 1 1 1 0
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 0 1 0 1 0
	// 0 0 0 1 1 0 1 1
	// 0 0 0 0 0 0 0 0
	0x00, 0x0e, 0x11, 0x11, 0x11, 0x0a, 0x1b, 0x00,

	// micro
	// 0 0 0 0 0 0 0 0
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 0 0 0 1
	// 0 0 0 1 0 0 1 1
	// 0 0 0 1 1 1 0 1
	// 0 0 0 1 0 0 0 0
	// 0 0 0 1 0 0 0 0
        0x00, 0x11, 0x11, 0x11, 0x13, 0x1d, 0x10, 0x10,
};

// enable pulse (takes 10 us)
void LCD_EnablePulse()
{
	EN_ON;		// enable
	waitus(7);	// delay 7 us
	EN_OFF;		// disable
	waitus(3);	// delay 3 us
}

// output lower nibble of data byte (takes 10 us)
void LCD_OutNibble(u8 data)
{
	PORTC = (PORTC & 0xf0) | (data & 0xf);
	LCD_EnablePulse(); // enable pulse
}

// output data (takes 80 us)
void LCD_OutByte(u8 data)
{
	LCD_OutNibble(data >> 4); // output higher nibble of data byte
	LCD_OutNibble(data); // output lower nibble of data byte
	waitus(60);	// delay 60 us
}

// write data byte (takes 80 us)
void LCD_WriteData(u8 data)
{
	DATA_MODE;	// switch to data mode
	LCD_OutByte(data); // output data byte
}

// write command byte (takes 80 us, some commands need additional waiting)
void LCD_WriteCmd(u8 cmd)
{
	CMD_MODE;	// switch to command mode
	LCD_OutByte(cmd); // output command byte
}

// clear display (requires 1.53 ms, takes 3 ms)
void LCD_Clear()
{
	LCD_WriteCmd(LCDCMD_CLEAR); // clear display
	waitms(3);	// delay 3 ms
}

// return to home position (requires 1.53 ms, takes 3 ms)
void LCD_Home()
{
	LCD_WriteCmd(LCDCMD_HOME); // return to home position
	waitms(3);	// delay 3 ms
}

// set entry mode: right=shift cursor right, shift=shift entire display (takes 80 us)
void LCD_Entry(Bool right, Bool shift)
{
	LCD_WriteCmd(LCDCMD_MOVE +
		(right ? B1 : 0) +
		(shift ? B0 : 0));
}

// set display on: disp=display on, cursor=cursor on, blink=blinking on (takes 80 us)
void LCD_DispOn(Bool disp, Bool cursor, Bool blink)
{
	LCD_WriteCmd(LCDCMD_DISPLAY +
		(disp ? B2 : 0) +
		(cursor ? B1 : 0) +
		(blink ? B0 : 0));
}

// set cursor shift: shift=cursor shift, right=shift to right (takes 80 us)
void LCD_Shift(Bool shift, Bool right)
{
	LCD_WriteCmd(LCDCMD_SHIFT +
		(shift ? B3 : 0) +
		(right ? B2 : 0));
}

// set CGRAM (character generator) address (takes 80 us)
void LCD_CGRAM(u8 addr)
{
	LCD_WriteCmd(LCDCMD_CGRAM + addr);
}

// set DDRAM (display data) address (takes 80 us)
void LCD_DDRAM(u8 addr)
{
	LCD_WriteCmd(LCDCMD_DDRAM + addr);
}

// initialize LCD display
void LCD_Init()
{
	// set port pins to output mode
	DDRC |= B0+B1+B2+B3;
	DDRB |= B0+B1;

	// disable output
	EN_OFF;

	// wait for more than 20 ms for LCD power up
	waitms(100);

	// byte synchronisation to new interface
	CMD_MODE; // set command mode

	// byte synchronisation using 8-bit interface
	LCD_OutNibble((LCDCMD_SETUP + B4) >> 4); // switch to 8 bit interface
	waitms(5); // delay 5 ms
	LCD_OutNibble((LCDCMD_SETUP + B4) >> 4); // switch to 8 bit interface
	waitms(2); // delay 2 ms
	LCD_OutNibble((LCDCMD_SETUP + B4) >> 4); // switch to 8 bit interface
	waitms(1); // delay 1 ms

	// switch to 4-bit interface
	LCD_OutNibble(LCDCMD_SETUP >> 4); // switch to 4 bit interface
	waitms(5); // delay 5 ms

	// set 4-bit interface, 2 rows, font 5x7
	LCD_WriteCmd(LCDCMD_SETUP + B3);

	// display off, cursor off, blinking off
	LCD_DispOn(False, False, False);

	// clear display, home cursor
	LCD_Clear();

	// auto move cursor right
	LCD_Entry(True, False);

	// display on, cursor off, blinking off
	LCD_DispOn(True, False, False);

	// load custom characters	
	s8 i, b;
	const u8* s = Font5x7;
	LCD_CGRAM(0);
	for (i = 0; i < CHAR_NUM*8; i++)
	{
		b = eeprom_read_byte(s);
		LCD_WriteData(b);
		s++;
	}

	// set cursor home position
	LCD_Home();
}
