
// ****************************************************************************
//                                 
//                 Check Resistor and Electrolytic Capacitor
//
// ****************************************************************************

#include "include.h"

// value is ESR
Bool ESRMode = False;

// capacitor value is in 0.1 pF scale (or in 10 pF scale otherwise)
Bool CapHiRes = False;

// resistor measured value (scale = -1, in 0.1 Ohm)
s32 ResVal = RMAX_VAL;

// tare R
s32 RTare = R_TARE;

// tare C
s32 CTare = C_TARE;

// resistor history
FILTER(ResHist);

// capacitor measured value (scale = -11, in 10 pF, range up to 20 mF)
s32 CapVal = CMAX_VAL;

// capacitor history
FILTER(CapHist);

// test resistor index
enum {
	R1_INX_SLOW = 0,
	R2_INX_SLOW,
	R3_INX_SLOW,
	R4_INX_SLOW,
	R5_INX_SLOW,
	R6_INX_SLOW,
	R7_INX_SLOW,
	R8_INX_SLOW,
	R8B_INX_SLOW,

	R1_INX_FAST,
	R2_INX_FAST,
	R3_INX_FAST,
	R4_INX_FAST,
	R5_INX_FAST,
	R6_INX_FAST,
	R7_INX_FAST,
	R8_INX_FAST,
	R8B_INX_FAST,

	RX_INX_NUM
};

// test resistor descriptor (size 12 bytes)
typedef struct {
	u8	speed;		// speed selector (2, 7)
	u8	resinx;		// resistor index (R1..R8)
	u16	mul;		// result multiplier (1, 2, 20, 200, 2000)
	u16	samples;	// number of samples
	u16	resval;		// resistor value
	u32	k;		// "k" coefficient = Vcc + some voltage bias (max. approx. 277000)
} sResDesc;

// How to determine resval and mul: At first, use resval=(Rn_VAL+R0_VAL)*10 and mul=1. Multiply resval by max. result (360000).
//    If result is higher than DWORD, divide it and multiply mul by the same value to compensate.
//    (be aware, R1_VAL R2_VAL and R0_VAL are in 0.1 Ohms)

// How to determine "k": measure high values of resistor and add some bias, if result is high.

// test resistor table (size 204 bytes)
const sResDesc ResDescTab[RX_INX_NUM] PROGMEM = {
 { ADC_SPEED,      R1, 1,     ADC_SAMPLES,      R1_VAL+R0_VAL,               (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*80,            },
 { ADC_SPEED,      R2, 1,     ADC_SAMPLES,      R2_VAL+R0_VAL,               (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*25,            },
 { ADC_SPEED,      R3, 5,     ADC_SAMPLES,      (R3_VAL+R0_VAL/10)*2,        (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*20,            },
 { ADC_SPEED,      R4, 20,    ADC_SAMPLES,      (R4_VAL+R0_VAL/10+1)/2,      (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*16,            },
 { ADC_SPEED,      R5, 100,   ADC_SAMPLES,      (R5_VAL+R0_VAL/10+5)/10,     (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*12,            },
 { ADC_SPEED,      R6, 500,   ADC_SAMPLES,      (R6_VAL+R0_VAL/10+25)/50,    (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*10,            },
 { ADC_SPEED,      R7, 2000,  ADC_SAMPLES,      (R7_VAL+R0_VAL/10+100)/200,  (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*10,            },
 { ADC_SPEED,      R8, 10000, ADC_SAMPLES,      (R8_VAL+R0_VAL/10+500)/1000, (u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*10,            },
 { ADC_SPEED,      R8, 10000, ADC_SAMPLES,      (R8B_VAL+R0_VAL/10+500)/1000,(u32)ADC_SAMPLES*1023+ADC_SAMPLES/10*1,             },

 { ADC_SPEED_FAST, R1, 1,     ADC_SAMPLES_FAST, R1_VAL+R0_VAL,               (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*105, },
 { ADC_SPEED_FAST, R2, 1,     ADC_SAMPLES_FAST, R2_VAL+R0_VAL,               (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*30,  },
 { ADC_SPEED_FAST, R3, 5,     ADC_SAMPLES_FAST, (R3_VAL+R0_VAL/10)*2,        (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*20,  },
 { ADC_SPEED_FAST, R4, 20,    ADC_SAMPLES_FAST, (R4_VAL+R0_VAL/10+1)/2,      (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*16,  },
 { ADC_SPEED_FAST, R5, 100,   ADC_SAMPLES_FAST, (R5_VAL+R0_VAL/10+5)/10,     (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*12,  },
 { ADC_SPEED_FAST, R6, 500,   ADC_SAMPLES_FAST, (R6_VAL+R0_VAL/10+25)/50,    (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*10,  },
 { ADC_SPEED_FAST, R7, 2000,  ADC_SAMPLES_FAST, (R7_VAL+R0_VAL/10+100)/200,  (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*10,  },
 { ADC_SPEED_FAST, R8, 10000, ADC_SAMPLES_FAST, (R8_VAL+R0_VAL/10+500)/1000, (u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*10,  },
 { ADC_SPEED_FAST, R8, 10000, ADC_SAMPLES_FAST, (R8B_VAL+R0_VAL/10+500)/1000,(u32)ADC_SAMPLES_FAST*1023+ADC_SAMPLES_FAST/10*0,   },
};

// resistor boundaries (size 8 bytes)
typedef struct {
	u8	r;		// boundary resistor index (R1_INX..R8_INX)
	u8 	r1;		// resistor 1 index (R1_INX..R8_INX)
	u8	r2;		// resistor 2 index (R1_INX..R8_INX)
	u8	w;		// resistor 1 weight (1, 2, 3, 4)
	s32	val;		// boundary resistor value (*10)
} sResBound;

#define RESBOUND_NUM 32

// resistor boundaries (size 264 bytes)
//  Two neighbouring resistors have relation 4.7 times:
//   weight 4, R1*4 + R2*0: R1_VAL*4.7^(1/8)=R1_VAL*1.21
//   weight 3, R1*3 + R2*1: R1_VAL*4.7^(3/8)=R1_VAL*1.79
//   weight 2, R1*2 + R2*2: R1_VAL*4.7^(5/8)=R1_VAL*2.63
//   weight 1, R1*1 + R2*3: R1_VAL*4.7^(7/8)=R1_VAL*3.87
const sResBound ResBoundTab[RESBOUND_NUM+1] EEMEM = {
	{ R1_INX, R1_INX_SLOW, R1_INX_SLOW, 2, R1_VAL*121/100 }, // R1
	{ R1_INX, R1_INX_SLOW, R2_INX_SLOW, 3, R1_VAL*179/100 }, // R1+R2
	{ R1_INX, R1_INX_SLOW, R2_INX_SLOW, 2, R1_VAL*263/100 }, // R1+R2
	{ R2_INX, R1_INX_SLOW, R2_INX_SLOW, 1, R1_VAL*387/100 }, // R1+R2

	{ R2_INX, R2_INX_SLOW, R2_INX_SLOW, 2, R2_VAL*121/100 }, // R2
	{ R2_INX, R2_INX_SLOW, R3_INX_SLOW, 3, R2_VAL*179/100 }, // R2+R3
	{ R2_INX, R2_INX_SLOW, R3_INX_SLOW, 2, R2_VAL*263/100 }, // R2+R3
	{ R3_INX, R2_INX_SLOW, R3_INX_SLOW, 1, R2_VAL*387/100 }, // R2+R3

	{ R3_INX, R3_INX_SLOW, R3_INX_SLOW, 2, R3_VAL*121/10 }, // R3
	{ R3_INX, R3_INX_SLOW, R4_INX_SLOW, 3, R3_VAL*179/10 }, // R3+R4
	{ R3_INX, R3_INX_SLOW, R4_INX_SLOW, 2, R3_VAL*263/10 }, // R3+R4
	{ R4_INX, R3_INX_SLOW, R4_INX_SLOW, 1, R3_VAL*387/10 }, // R3+R4

	{ R4_INX, R4_INX_SLOW, R4_INX_SLOW, 2, R4_VAL*121/10 }, // R4
	{ R4_INX, R4_INX_SLOW, R5_INX_SLOW, 3, R4_VAL*179/10 }, // R4+R5
	{ R4_INX, R4_INX_SLOW, R5_INX_SLOW, 2, R4_VAL*263/10 }, // R4+R5
	{ R5_INX, R4_INX_SLOW, R5_INX_SLOW, 1, R4_VAL*387/10 }, // R4+R5

	{ R5_INX, R5_INX_SLOW, R5_INX_SLOW, 2, R5_VAL*121/10 }, // R5
	{ R5_INX, R5_INX_SLOW, R6_INX_SLOW, 3, R5_VAL*179/10 }, // R5+R6
	{ R5_INX, R5_INX_SLOW, R6_INX_SLOW, 2, R5_VAL*263/10 }, // R5+R6
	{ R6_INX, R5_INX_SLOW, R6_INX_SLOW, 1, R5_VAL*387/10 }, // R5+R6

	{ R6_INX, R6_INX_SLOW, R6_INX_SLOW, 2, R6_VAL*121/10 }, // R6
	{ R6_INX, R6_INX_SLOW, R7_INX_SLOW, 3, R6_VAL*179/10 }, // R6+R7
	{ R6_INX, R6_INX_SLOW, R7_INX_SLOW, 2, R6_VAL*263/10 }, // R6+R7
	{ R7_INX, R6_INX_SLOW, R7_INX_SLOW, 1, R6_VAL*387/10 }, // R6+R7

	{ R7_INX, R7_INX_SLOW, R7_INX_SLOW, 2, R7_VAL*121/10 }, // R7
	{ R7_INX, R7_INX_SLOW, R8_INX_SLOW, 3, R7_VAL*179/10 }, // R7+R8
	{ R7_INX, R7_INX_SLOW, R8_INX_SLOW, 2, R7_VAL*263/10 }, // R7+R8
	{ R8_INX, R7_INX_SLOW, R8_INX_SLOW, 1, R7_VAL*387/10 }, // R7+R8

	{ R8_INX, R8_INX_SLOW, R8_INX_SLOW, 2, R8_VAL*121/10 }, // R8
	{ R8_INX, R8_INX_SLOW, R8B_INX_SLOW, 3, R8_VAL*179/10 }, // R8+R8B
	{ R8_INX, R8_INX_SLOW, R8B_INX_SLOW, 2, R8_VAL*263/10 }, // R8+R8B
	{ R8B_INX, R8_INX_SLOW, R8B_INX_SLOW, 1, R8_VAL*387/10 }, // R8+R8B

	{ R8B_INX, R8B_INX_SLOW, R8B_INX_SLOW, 2, R8_VAL*470/10 }, // R8B
};

// table of resistor values (size 36 bytes)
const u32 ResValTab[RX_NUM] EEMEM = {
	R1_VAL/10+R0_VAL/10,
	R2_VAL/10+R0_VAL/10,
	R3_VAL+R0_VAL/10,
	R4_VAL+R0_VAL/10,
	R5_VAL+R0_VAL/10,
	R6_VAL+R0_VAL/10,
	R7_VAL+R0_VAL/10,
	R8_VAL+R0_VAL/10,
	R8B_VAL+R0_VAL/10,
};

// table of resistor indices (size 9 bytes)
const u8 ResInxTab[RX_NUM] EEMEM = { R1, R2, R3, R4, R5, R6, R7, R8, R8 };

// capacitor boundaries (size 8 bytes)
typedef struct {
	u8 	rA;		// resistor A index (R1..R8)
	u8	rB;		// resistor B index (R1..R8)
	u16	res; 		// ... reserved, unused
	s32	val;		// boundary capacitor value (in multiply of 10 pF)
} sCapBound;

#define CAPBOUND_NUM 14

// capacitor boundaries (size 120 bytes)
const sCapBound CapBoundTab[CAPBOUND_NUM+1] PROGMEM = {
	{ R8_INX, R8_INX, 0, 100L }, // 1nF
	{ R8_INX, R7_INX, 0, 300L }, // 3nF
	{ R7_INX, R7_INX, 0, 1000L }, // 10nF
	{ R7_INX, R6_INX, 0, 3000L }, // 30nF
	{ R6_INX, R6_INX, 0, 10000L }, // 100nF
	{ R6_INX, R5_INX, 0, 30000L }, // 300nF
	{ R5_INX, R5_INX, 0, 100000L }, // 1uF
	{ R5_INX, R4_INX, 0, 300000L }, // 3uF
	{ R4_INX, R4_INX, 0, 1000000L }, // 10uF
	{ R4_INX, R3_INX, 0, 3000000L }, // 30uF
	{ R3_INX, R3_INX, 0, 10000000L }, // 100uF
	{ R3_INX, R2_INX, 0, 30000000L }, // 300uF
	{ R2_INX, R2_INX, 0, 100000000L }, // 1mF
	{ R2_INX, R1_INX, 0, 300000000L }, // 3mF

	{ R1_INX, R1_INX, 0, 1000000000L }, // 10mF
};

// measure ESR resistance with one resistor (takes 20 ms)
// Returns resistance in 0.1 ohms, or RMAX_VAL if charging is too fast
s32 RC_CheckESR()
{
	// select speed (14 us per ADC_Sample)
	ADC_Speed(ADC_SPEED_FAST);

	// prepare operands
	u16 i;
	s16 v, v1, v2, v3;
	s32 r;
	s32 a = 0; // resistor accumulator

	SetRes(R1, 0); // set R1 to LOW, discharge capacitor

#define ESR_SAMPLES 120 // number of loops (20 ms)

	// loop (one step approx. 50 + 3*14 + 75 = 167 us)
	for (i = ESR_SAMPLES; i > 0; i--)
	{
		// set R1 to LOW, discharge capacitor
		R1_OUT();

		// delay to discharge capacitor
		waitus(50);

		// switch all resistors OFF
		R1_IN();

		// load sample at OFF state; takes 14 us
		v1 = ADC_Sample();

		// switch R1 to HIGH, charge capacitor
		R1_HIGH();
		R1_OUT();

		// load sample at HIGH state; takes 14 us
		v2 = ADC_Sample();

		// switch all resistors OFF
		R1_IN();
		R1_LOW();

		// load sample at OFF state; takes 14 us
		v3 = ADC_Sample();

		// check overflow, capacitor is too small or ESR is too high
		r = v2 - v1;
		if (((v3 - v1) > 700) || (r > 700)) return RMAX_VAL;		

		// calculate resistor value (takes approx. 75 us on 16 MHz, 100 us on 12 MHz, 160 us on 7.3728 MHz)
		// I = (Vcc - v2) / (R1 + R0)
		// R = (v2 - v1) / I = (v2 - v1) * (R1 + R0) / (Vcc - v2)
		v = (1023 + 2) - v2; // value 2..1025
		if (r < 0) r = 0; // value 0..700
		r = (r * (R1_VAL + R0_VAL) + v/2) / v; // result in 0.1 Ohm, value 0..800 kOhm

		// add to accumulator
		a += r;
	}

	// set all resistors OFF
	SetResOff();

	// get ESR value
	a = (a + ESR_SAMPLES/2) / ESR_SAMPLES;

	// limit to 100 Ohm
	if (a >= 1000L) return RMAX_VAL;
	return a - RTare;
}

// check one type of resistor (returns resistance in 0.1 ohms, takes 20..25 ms or 200..5000 us)
//   rinx = resistor index

// I = Vcc / (R + Rn)
// v = I * R = Vcc / (R + Rn) * R; v*R + v*Rn = Vcc*R
// R = v*Rn/(Vcc - v)
s32 RC_CheckResX(u8 rinx)
{
	// get descriptor
	const sResDesc* rd = &ResDescTab[rinx];

	// switch to UP direction
	//SetRes(0, eeprom_read_byte(&rd->resinx));
	SetRes(0, pgm_read_byte(&rd->resinx));

	// wait for stabilisation
	if ((rinx == R6_INX_SLOW) || (rinx == R6_INX_FAST))
		waitus(150);

	if ((rinx == R7_INX_SLOW) || (rinx == R7_INX_FAST))
		waitus(500);

	if ((rinx == R8_INX_SLOW) || (rinx == R8_INX_FAST))
		waitus(2000);

	if ((rinx == R8B_INX_SLOW) || (rinx == R8B_INX_FAST))
		waitus(5000);

	// load samples (takes 20 ms or 200 us)
	//u32 v = ADC_Read(eeprom_read_word((const uint16_t*)&rd->samples)); // v = 0..358050
	u32 v = ADC_Read(pgm_read_word((const uint16_t*)&rd->samples)); // v = 0..358050

	// switch all resistors OFF
	SetResOff();

	// prepare divisor, k = Vcc - v (value 15..358100)
	//u32 k = eeprom_read_dword(&rd->k) - v;
	u32 k = pgm_read_dword(&rd->k) - v;

	// multiply by resistor value, v = v*Rn
	//v *= eeprom_read_word((const uint16_t*)&rd->resval); // v = 0..4G
	v *= pgm_read_word((const uint16_t*)&rd->resval); // v = 0..4G

	// calculate resistance R = (v*Rn) / (Vcc - v)
	//u32 r = (u32)(v + k/2) / k * eeprom_read_word((const uint16_t*)&rd->mul);
	u32 r = (u32)(v + k/2) / k * pgm_read_word((const uint16_t*)&rd->mul);

	if (r >= (u32)RMAX_VAL) return RMAX_VAL;
	return r - RTare;
}

// check simple resistor (takes 50..60 ms)
void RC_CheckRes()
{
#ifdef DO_CHECKRES	// check one specific reference resistor
	ADC_Speed(ADC_SPEED);

	int roff = 0;
#ifdef DO_CHECKRESFAST // flag - do DO_CHECKRES in fast mode 
	roff = R1_INX_FAST - R1_INX_SLOW;
#endif

#if DO_CHECKRES==1
	s32 rrr = RC_CheckResX(R1_INX_SLOW+roff);
#elif DO_CHECKRES==2
	s32 rrr = RC_CheckResX(R2_INX_SLOW+roff);
#elif DO_CHECKRES==3
	s32 rrr = RC_CheckResX(R3_INX_SLOW+roff);
#elif DO_CHECKRES==4
	s32 rrr = RC_CheckResX(R4_INX_SLOW+roff);
#elif DO_CHECKRES==5
	s32 rrr = RC_CheckResX(R5_INX_SLOW+roff);
#elif DO_CHECKRES==6
	s32 rrr = RC_CheckResX(R6_INX_SLOW+roff);
#elif DO_CHECKRES==7
	s32 rrr = RC_CheckResX(R7_INX_SLOW+roff);
#elif DO_CHECKRES==8
	s32 rrr = RC_CheckResX(R8_INX_SLOW+roff);
#else
	s32 rrr = RC_CheckResX(R8B_INX_SLOW+roff);
#endif
	ResVal = FilterAdd(&ResHist, rrr);
	return;
#endif // DO_CHECKRES

	// select fast speed
	ADC_Speed(ADC_SPEED_FAST);

	// fast get resistors (takes 9.5 ms)
	s32 rr[RX_NUM];
	rr[R1_INX] = RC_CheckResX(R1_INX_FAST); // 200 us
	rr[R2_INX] = RC_CheckResX(R2_INX_FAST); // 200 us
	rr[R3_INX] = RC_CheckResX(R3_INX_FAST); // 200 us
	rr[R4_INX] = RC_CheckResX(R4_INX_FAST); // 200 us
	rr[R5_INX] = RC_CheckResX(R5_INX_FAST); // 200 us
	rr[R6_INX] = RC_CheckResX(R6_INX_FAST); // 350 us
	rr[R7_INX] = RC_CheckResX(R7_INX_FAST); // 700 us
	rr[R8_INX] = RC_CheckResX(R8_INX_FAST); // 2200 us
	rr[R8B_INX] = RC_CheckResX(R8B_INX_FAST); // 5200 us

	// select slow speed
	ADC_Speed(ADC_SPEED);

	// find boundary
	const sResBound* rb = ResBoundTab;
	u8 i;
	for (i = RESBOUND_NUM; i > 0; i--)
	{
		if (rr[eeprom_read_byte(&rb->r)] < (s32)eeprom_read_dword((u32*)&rb->val)) break;
		rb++;
	}

	// interpolate (takes 40..50 ms)
	u8 r1k = eeprom_read_byte(&rb->r1);
	u8 r2k = eeprom_read_byte(&rb->r2);
	s8 w = eeprom_read_byte(&rb->w);

	s32 r, rr1, rr2;

#ifdef DO_CHECKRESFAST // flag - use FAST mode
	rr1 = rr[r1k];
	rr2 = rr[r2k];
#else
	rr1 = RC_CheckResX(r1k);
	rr2 = RC_CheckResX(r2k);
#endif
	if (rr1 == RMAX_VAL)
		r = rr2;
	else
		if (rr2 == RMAX_VAL)
			r = rr1;
		else
		{
			if ((rr1 >= 0) && (rr2 >= 0))
				r = (s32)((u32)((u32)rr1*(u8)w + (u32)rr2*(u8)(4-w) + 2)/4);
			else
				r = (rr1*w + rr2*(4-w))/4;
		}

	// add resistor to resistor history
	ResVal = FilterAdd(&ResHist, r);
}

// detect capacitor (returns True if capacitor has been detected) (takes 5 ms)
Bool RC_DetCap()
{
	// select fast speed
	ADC_Speed(ADC_SPEED_FAST);

	// fast charge using R1 (100) with delay 2 ms
	SetRes(0, R1);
	waitms(2);

	// check voltage, discharge using R5 (47K) with some delay
	cli();
	SetRes(R5, 0);
	waitus(10);
	u16 v = ADC_Sample();
	sei();

	// switch resistors OFF
	SetResOff();

	// check (big capacitors may need more loops to charge)
	return (v > 8);
}

// check capacitor using 1 resistor (takes 20 ms, or up to 520 ms if discharge)
// - fast mode takes 1/10 of time
s32 RC_CheckCap1(u8 rinx, u32 rval, Bool fast)
{
	// sum
	u32 sum = 0;
	u16 num = 0;
	s32 tim = fast ? 2000 : 20000;

	// set speed to fast
	ADC_Speed(ADC_SPEED_FAST);
	Bool slow = !fast && (rinx < R7);

	// while some time left
	u16 i, limit, v1, v2, v0;
	s16 v1b, dif;
	s32 t;
	u32 c, tim1, tim2, ri, r, v;
	while (tim > 50)
	{
		// set speed to fast
		if (slow) ADC_Speed(ADC_SPEED_FAST);

		// discharging by all resistors down
		SetRes(RALL, 0);

		// prepare
		i = 2500;
		limit = 150;
		if (!fast)
		{
			i = 25000;
			limit = 70;
		}

		// discharge (max. 500 ms or 50 ms)
		for (; i > 0; i--)
		{
			// check voltage (takes 14 us)
			if (ADC_Sample() < limit) break;
		}

		// set all resistors OFF
		SetResOff();

		// prepare
		i = tim/14; // time on one ADC_Sample step - correct if needed!
		limit = 50;
		if (!fast) limit = 100;

		// set speed to slow
		if (slow) ADC_Speed(ADC_SPEED);

		// get start voltage
		v1 = ADC_Sample();

		// prepare stop voltage
		v1b = v1 + limit;

		// set speed to fast
		if (slow) ADC_Speed(ADC_SPEED_FAST);

		// get start time (takes 3 us, resolution 4 us)
		tim1 = GetTime();

		// start charge (takes 15 us)
		SetRes(0, rinx);

		// wait for stop voltage (max. 40 ms or 4 ms, takes 14 us per sample)
		for (; i > 0; i--) if ((s16)ADC_Sample() >= v1b) break;

		// stop discharging (takes 15 us) (do not use SetResOff, to ensure the same timings)
		SetRes(0, 0);

		// get stop time (takes 3 us, resolution 4 us)
		tim2 = GetTime();

		// set speed to slow
		if (slow) ADC_Speed(ADC_SPEED);

		// get stop voltage
		v2 = ADC_Sample();

		// time difference (value 1..20000)
		t = tim2 - tim1 - GETTIME_US - SETRES_US;
		if (t < 1) t = 1;
		tim -= t + 50;

		// voltage difference (value 0..150)
		dif = v2 - v1;
		if (dif < 0) dif = -dif;
		if (dif < 3) return CMAX_VAL;

		// internal resistor (in 0.1 Ohm)
		ri = ResVal;
		if (ri >= 1000) ri = 0;

		// charging resistors (value 100..4M7)
		r = rval + ri/10;

		// resistor * voltage difference (value 100..500M)
		r *= dif;

		// high resolution
		if (CapHiRes) r /= 100;

		// middle voltage * 2 (value 0..2046)
		v0 = v1 + v2;

		// voltage on resistors (value 2..2048)
		v0 = (1023Ul + 1)*2 - v0;
		if (v0 < 2) v0 = 2;

		// voltage * time * 10 (value 100..400M)
		v = v0 * t * 5;

		// capacity in 10 pF 
		if (v >= 400000000)
			c = (v + r/2) / r * 10000;
		else if (v >= 40000000)
			c = (v*10 + r/2) / r * 1000;
		else if (v >= 4000000)
			c = (v*100 + r/2) / r * 100;
		else
			c = (v*1000 + r/2) / r * 10;

		// add to sum
		sum += c;
		num++;
	}

	// average
	c = (sum + num/2)/num;
	if (c > CMAX_VAL) return CMAX_VAL;

	// tare
	sum = CapHiRes ? CTare : (CTare/100);

	return c - sum;
}

// check capacitor (takes 50 ms, or up to 550 ms if discharging)
void RC_CheckCap()
{
#ifdef DO_CHECKCAP	// check one specific reference resistor to measure capacitors (set 1..8 to check R1..R8)

#ifdef DO_CHECKCAPFAST // flag - do DO_CHECKCAP in fast mode 
	Bool fast = True;
#else
	Bool fast = False;
#endif

#if DO_CHECKCAP == 1
	s32 cc = RC_CheckCap1(R1, R1_VAL/10+R0_VAL/10, fast);
#elif DO_CHECKCAP==2
	s32 cc = RC_CheckCap1(R2, R2_VAL/10+R0_VAL/10, fast);
#elif DO_CHECKCAP==3
	s32 cc = RC_CheckCap1(R3, R3_VAL+R0_VAL/10, fast);
#elif DO_CHECKCAP==4
	s32 cc = RC_CheckCap1(R4, R4_VAL+R0_VAL/10, fast);
#elif DO_CHECKCAP==5
	s32 cc = RC_CheckCap1(R5, R5_VAL+R0_VAL/10, fast);
#elif DO_CHECKCAP==6
	s32 cc = RC_CheckCap1(R6, R6_VAL+R0_VAL/10, fast);
#elif DO_CHECKCAP==7
	s32 cc = RC_CheckCap1(R7, R7_VAL+R0_VAL/10, fast);
#else
	s32 cc = RC_CheckCap1(R8, R8_VAL+R0_VAL/10, fast);
#endif
	CapVal = FilterAdd(&CapHist, cc);
	return;
#endif

	// detect capacitor fast (takes 2 or 4 ms)
	s32 c = RC_CheckCap1(R4, R4_VAL+R0_VAL/10, True);
	s32 cx = c;
	if (CapHiRes && (c != CMAX_VAL)) c /= 100;
	if (c <= 1000L) // 10 nF
	{
		CapHiRes = True;
		c = RC_CheckCap1(R8, R8_VAL+R0_VAL/10, True);
		cx = c;
		if (c != CMAX_VAL) c /= 100;
	}
	else
	{
		if (c >= 100000L) // 1 uF
		{
			CapHiRes = False;
			c = RC_CheckCap1(R1, R1_VAL/10+R0_VAL/10, True);
			cx = c;
		}
	}

#ifdef DO_CHECKCAPFAST // flag - result in fast mode 
	CapVal = FilterAdd(&CapHist, cx);
	return;
#endif

	// find boundary
	const sCapBound* cb = CapBoundTab;
	u8 i;
	for (i = CAPBOUND_NUM; i > 0; i--)
	{
		if (c < (s32)pgm_read_dword((u32*)&cb->val)) break;
		cb++;
	}

	// interpolate (takes 2*20=40 ms)
	u8 rAk = pgm_read_byte(&cb->rA);
	u8 rBk = pgm_read_byte(&cb->rB);
	s32 c1 = RC_CheckCap1(eeprom_read_byte(&ResInxTab[rAk]), eeprom_read_dword(&ResValTab[rAk]), False);
	s32 c2 = RC_CheckCap1(eeprom_read_byte(&ResInxTab[rBk]), eeprom_read_dword(&ResValTab[rBk]), False);
	if (c1 == CMAX_VAL)
		c = c2;
	else
		if (c2 == CMAX_VAL)
			c = c1;
		else
		{
			if ((c1 >= 0) && (c2 >= 0))
				c = (u32)((u32)c1 + (u32)c2 + 1)/2;
			else
				c = (c1 + c2)/2;
		}

	// add capacitor to capacitor history
	CapVal = FilterAdd(&CapHist, c);
}

// check resistor (takes 65 ms)
void RC_Check()
{
	// select Vcc reference
	ADC_Sel(ADC_MUX_IN); //, ADC_REF_VCC);

	// detect capacitor (takes 5 ms)
	ESRMode = False;
	if (!RC_DetCap())
	{
		// check simple resistor (takes 50..60 ms)
		RC_CheckRes();
	}
	else
	{
		// check ESR (takes 20 ms)
		s32 r = RC_CheckESR();
		if (r == RMAX_VAL)
			RC_CheckRes(); // high resistance detected as capacitor
		else
		{
			ResVal = FilterAdd(&ResHist, r); // add to history
			ESRMode = True;
		}
	}
}
