
// ****************************************************************************
//                                 
//                          System Clock (using Timer)
//
// ****************************************************************************

// timer prescaler
//   1 ... 1
//   2 ... 1/8
//   3 ... 1/64
//   4 ... 1/256
//   5 ... 1/1024

// To check how much interrupts load processor - use DO_CLOCKTEST, timers should be 1.000.

// prescaler 1: 1, period = 1 / (F_CPU / 256):
// Interrupts load processor by 60%
//   16 us on 16 MHz (62.50 kHz), resolution 0.062 us
//   21.3333 us on 12 MHz (46.88 kHz), resolution 0.083 us
//   32 us on 8 MHz (331.25 kHz), resolution 0.125 us
//   34.7222 us on 7.372800 MHz (28.80 kHz), resolution 0.135 us
//   64 us on 4 MHz (15.63 kHz), resolution 0.25 us
//   128 us on 2 MHz (7.812 kHz), resolution 0.5 us
//   256 us on 1 MHz (3.906 kHz), resolution 1 us
//   7812.5 us on 32768 Hz (128 Hz), resolution 30.52 us

// prescaler 2: 1/8, period = 1 / (F_CPU / (8*256)):
// Interrupts load processor by 6.9%
//   128 us on 16 MHz (7.813 kHz), resolution 0.5 us
//   170.6667 us on 12 MHz (5.859 kHz), resolution 0.7 us
//   256 us on 8 MHz (3.906 kHz), resolution 1 us
//   277.7778 us on 7.372800 MHz (3.6 kHz), resolution 1.1 us
//   512 us on 4 MHz (1.953 kHz), resolution 2 us
//   1024 us on 2 MHz (976.6 Hz), resolution 4 us
//   2048 us on 1 MHz (488.3 Hz), resolution 8 us
//   62500 us on 32768 Hz (16 Hz), resolution 244.1 us

// prescaler 3: 1/64, period = 1 / (F_CPU / (64*256)):
// Interrupts load processor by 0.8%
//   1024 us on 16 MHz (976.6 Hz), resolution 4 us
//   1365.3333 us on 12 MHz (732.4 Hz), resolution 5.3 us
//   2048 us on 8 MHz (488.3 Hz), resolution 8 us
//   2222.2222 us on 7.372800 MHz (450 Hz), resolution 8.7 us
//   4096 us on 4 MHz (244.1 Hz), resolution 16 us
//   8192 us on 2 MHz (122.1 Hz), resolution 32 us
//   16384 us on 1 MHz (61 Hz), resolution 64 us
//   500000 us on 32768 Hz (2 Hz), resolution 1.953 ms

// prescaler 4: 1/256, period = 1 / (F_CPU / (256*256)):
// Interrupts load processor by 0.4%
//   4096 us on 16 MHz (244 Hz), resolution 16 us
//   5461.3334 us on 12 MHz (183 Hz), resolution 21.33 us
//   8192 us on 8 MHz (122 Hz), resolution 32 us
//   8888.8886 us on 7.372800 MHz (112 Hz), resolution 34.72 us
//   16384 us on 4 MHz (61 Hz), resolution 64 us
//   32768 us on 2 MHz (30 Hz), resolution 128 us
//   65536 us on 1 MHz (15 Hz), resolution 256 us
//   2000000 us on 32768 Hz (0.5 Hz), resolution 7.8 ms

#define SYSTIME_PRESCALER 3 // index of prescaler 1..5

// prescaler divider
#if SYSTIME_PRESCALER >= 3
#define SYSTIME_DIV (1 << (SYSTIME_PRESCALER*2))
#elif SYSTIME_PRESCALER == 2
#define SYSTIME_DIV 8
#else
#define SYSTIME_DIV 1
#endif

// interrupt frequency in Hz
#define SYSTIME_FREQ	((double)F_CPU / SYSTIME_DIV / 256)

// interrupt period in us
#define SYSTIME_TIME	(1000000UL / SYSTIME_FREQ + 5e-5)

// interrupt period HIGH (in us)
#define SYSTIME_HIGH	((u32)SYSTIME_TIME)

// interrupt period LOW (in 0.1 ns) - program can be optimised if SYSTIME_LOW = 0
#define SYSTIME_LOW	((u32)((SYSTIME_TIME - SYSTIME_HIGH)*10000))

// time resolution in ns
#define SYSTIME_RES	((u32)(SYSTIME_DIV / (double)F_CPU * 1000000000UL))

// do not use SYSTIME_LOW
#if (F_CPU == 1000000) || (F_CPU == 2000000) || (F_CPU == 4000000) || (F_CPU == 8000000) || (F_CPU == 16000000)
#define NO_SYSTIME_LOW
#endif

// initialize timer
void Timer_Init();

// prepare delta time of GetTime function in us (used in test functions)
// - takes 2.7 us on 16 MHz, 8.4 us on 12 MHz, 6.3 us on 8 MHz, 13.7 us on 7.3728 MHz,
//   14.4 us on 4 MHz, 32.4 us on 2 MHz, 72 us on 1 MHz
#if F_CPU >= 14000000
#define GETTIME_US 3 // 16 MHz
#elif F_CPU >= 10000000
#define GETTIME_US 8 // 12 MHz
#elif F_CPU >= 7600000
#define GETTIME_US 6 // 8 MHz
#elif F_CPU >= 6000000
#define GETTIME_US 14 // 7.3728 MHz
#elif F_CPU >= 3000000
#define GETTIME_US 14 // 4 MHz
#elif F_CPU >= 1500000
#define GETTIME_US 32 // 2 MHz
#elif F_CPU >= 300000
#define GETTIME_US 72 // 1 MHz
#endif

// get system time in number of us from restart, with period 71 minutes
// - Temporary disables and enables interrupts
u32 GetTime();
